#include "buffer.h"

/* * * Inline extern prototypes * * */

LSUP_Buffer *LSUP_buffer_new (const unsigned char *data, const size_t size);
LSUP_rc LSUP_buffer_init (
        LSUP_Buffer *buf, const size_t size, const unsigned char *data);
bool LSUP_buffer_eq (const LSUP_Buffer *buf1, const LSUP_Buffer *buf2);
int LSUP_buffer_cmp (const LSUP_Buffer *buf1, const LSUP_Buffer *buf2);
static inline char unescape_char (const char c);


/* * * API * * */

LSUP_rc
LSUP_buffer_init (
        LSUP_Buffer *buf, const size_t size, const unsigned char *data)
{
    // If size is zero, addr becomes NULL.
    unsigned char *tmp = realloc (buf->addr, size);
    if (UNLIKELY (size > 0 && tmp == NULL)) return LSUP_MEM_ERR;

    buf->addr = tmp;
    buf->size = size;

    if (data) memcpy (buf->addr, data, buf->size);

    return LSUP_OK;
}


int LSUP_utf8_encode (const uint32_t utf, unsigned char *out)
{
  if (utf <= 0x7F) {
    // Plain ASCII
    out[0] = (char) utf;
    out[1] = 0;
    return 1;
  }
  else if (utf <= 0x07FF) {
    // 2-byte unicode
    out[0] = (char) (((utf >> 6) & 0x1F) | 0xC0);
    out[1] = (char) (((utf >> 0) & 0x3F) | 0x80);
    out[2] = 0;
    return 2;
  }
  else if (utf <= 0xFFFF) {
    // 3-byte unicode
    out[0] = (char) (((utf >> 12) & 0x0F) | 0xE0);
    out[1] = (char) (((utf >>  6) & 0x3F) | 0x80);
    out[2] = (char) (((utf >>  0) & 0x3F) | 0x80);
    out[3] = 0;
    return 3;
  }
  else if (utf <= 0x10FFFF) {
    // 4-byte unicode
    out[0] = (char) (((utf >> 18) & 0x07) | 0xF0);
    out[1] = (char) (((utf >> 12) & 0x3F) | 0x80);
    out[2] = (char) (((utf >>  6) & 0x3F) | 0x80);
    out[3] = (char) (((utf >>  0) & 0x3F) | 0x80);
    out[4] = 0;
    return 4;
  }
  else {
    // error - use replacement character
    out[0] = (char) 0xEF;
    out[1] = (char) 0xBF;
    out[2] = (char) 0xBD;
    out[3] = 0;
    return 0;
  }
}


LSUP_Buffer *LSUP_buffer_from_str (const char *str)
{
    LSUP_Buffer *buf;
    MALLOC_GUARD (buf, NULL);
    buf->addr = malloc (strlen (str) + 1);
    buf->size = 1; // At least the NUL byte.

    size_t j = 0;
    unsigned char tmp_chr[5];
    for (size_t i = 0; i < strlen (str);) {
        if (str[i] == '\\') {
            i++; // Skip over '\\'

            // 4-hex sequence.
            if (str[i] == 'u') {
                i ++; // Skip over 'u'

                // Use tmp_chr to hold the hex string for the code point.
                memcpy(tmp_chr, str + i, sizeof (tmp_chr) - 1);
                tmp_chr[4] = '\0';

                uint32_t tmp_val = strtol ((char*)tmp_chr, NULL, 16);
                log_debug ("tmp_val: %d", tmp_val);

                // Reuse tmp_chr to hold the byte values for the code point.
                int nbytes = utf8_encode (tmp_val, tmp_chr);
                if (nbytes == 0)
                    log_error ("Error encoding code point: %u", tmp_val);

                // Copy bytes into destination.
                memcpy (buf->addr + j, tmp_chr, nbytes);
                /*
                log_trace (
                        "UC byte value: %x %x",
                        buf->addr[j], buf->addr[j + 1]);
                */

                j += nbytes;
                i += 4;

            // 8-hex sequence.
            } else if (str[i] == 'U') {
                i ++; // Skip over 'U'
                log_error ("UTF-16 sequence unescaping not yet implemented.");
                return NULL; // TODO encode UTF-16

            // Unescape other escaped characters.
            } else buf->addr[j++] = unescape_char(str[i++]);
        } else {
            // Copy ASCII char verbatim.
            buf->addr[j++] = str[i++];
        }
        buf->size ++;
    }

    unsigned char *tmp = realloc (buf->addr, j + 1); // Compact data.
    if (UNLIKELY (!tmp)) return NULL;
    buf->addr = tmp;
    buf->addr[j] = '\0';
    buf->size ++;

    return buf;
}




void LSUP_buffer_print (const LSUP_Buffer *buf)
{
    for (size_t i = 0; i < buf->size; i++) {
        char chr = ((char*)buf->addr)[i];
        if (isprint (chr)) {
            fputc (chr, stdout);
        } else {
            printf ("\\x%02x", chr);
        }
    }
    printf ("\n");
}


char *
LSUP_buffer_as_str (const LSUP_Buffer *buf)
{
    size_t i, str_size = 1; // terminating NUL

    // Calculate alloc size first.
    for (i = 0; i < buf->size; i++) {
        char chr = ((char*)buf->addr)[i];
        if (isprint (chr)) str_size ++;
        else str_size += 4; // 4 characters for ASCII representation (\xNN).
    }

    log_trace ("Byte buffer str size: %lu", str_size);

    char *cstr = malloc (str_size);

    size_t cur = 0; // Position in target string.
    for (i = 0; i < str_size - 1; i++) {
        char chr = ((char*)buf->addr)[i];

        if (isprint (chr)) {
            cstr[cur] = chr;
            cur ++;
        } else {
            sprintf (cstr + cur, "\\x%02x", chr);
            cur += 4;
        }
    }

    cstr[str_size - 1] = 0; // terminating NUL.

    return cstr;
}


void LSUP_buffer_done (LSUP_Buffer *buf)
{
    if (LIKELY (buf)) free (buf->addr);
}

void LSUP_buffer_free (LSUP_Buffer *buf)
{
    LSUP_buffer_done (buf);
    free (buf);
}


/*
 * Buffer triples.
 */

LSUP_BufferTriple *
LSUP_btriple_new(LSUP_Buffer *s, LSUP_Buffer *p, LSUP_Buffer *o)
{
    LSUP_BufferTriple *sspo = malloc (sizeof (*sspo));
    if (!sspo) return NULL;

    if (UNLIKELY (LSUP_btriple_init (sspo, s, p, o))) {
        free (sspo);
        return NULL;
    }

    return sspo;
}


LSUP_rc
LSUP_btriple_init (
        LSUP_BufferTriple *sspo, LSUP_Buffer *s, LSUP_Buffer *p, LSUP_Buffer *o)
{
    sspo->s = s;
    sspo->p = p;
    sspo->o = o;

    return LSUP_OK;
}


void
LSUP_btriple_done (LSUP_BufferTriple *sspo)
{
    if (UNLIKELY (!sspo)) return;

    LSUP_buffer_done (sspo->s);
    LSUP_buffer_done (sspo->p);
    LSUP_buffer_done (sspo->o);
}


void
LSUP_btriple_free (LSUP_BufferTriple *sspo)
{
    if (UNLIKELY (!sspo)) return;

    LSUP_buffer_free (sspo->s);
    LSUP_buffer_free (sspo->p);
    LSUP_buffer_free (sspo->o);

    free (sspo);
}


void
LSUP_btriple_free_shallow (LSUP_BufferTriple *sspo)
{
    if (UNLIKELY (!sspo)) return;

    sspo->s->addr = sspo->p->addr = sspo->o->addr = NULL;
    LSUP_btriple_free (sspo);
}


/*
 * Statics.
 */

static inline char unescape_char (const char c) {
    switch (c) {
        case 't': return '\t';
        case 'b': return '\b';
        case 'n': return '\n';
        case 'r': return '\r';
        case 'f': return '\f';
        default: return c;
    }
}


/* Extern inline prototypes. */

LSUP_Key LSUP_buffer_hash (const LSUP_Buffer *buf);
LSUP_Key LSUP_btriple_hash (const LSUP_BufferTriple *strp);
LSUP_Buffer *LSUP_btriple_pos (const LSUP_BufferTriple *trp, LSUP_TriplePos n);
